=begin

== Name

eachsynchronizer.rb - synchronize multiple objects' 'each' call.

Version 0.1.0

Copyright(c) 2001
Okada Jun
yun@nekome.net
http://www.nekome.net/


== Detail

  EachSynchronizer synchronize multiple objects, which have
 'each' method.

  First, create EachSynchronizer object to pass over target 
 objects tha have 'each' method.

  Next, call EachSynchronizer#each.

  Finally, multiple objects' 'each' method are called 
 synchoronously, and resulted arguments are passed to
 EachSynchronizers given block.

  Also, you are able to use useful shortcut, Array#synchoronous_each.

== Example

  a = [58, 38,  6, 97, 26, 86]
  b = "5\n19\n61\n60\n9\n30"
  c = 50..60
  prog = [a, b, c]
 
  result = []
  
  prog.synchronous_each do |args|
    result << args.collect{|n| n.to_i}.max
  end
  
  p result
   => [58, 51, 61, 97, 54, 86]

=end


require 'thread'


class EachSynchronizer
  include Enumerable

  class EndOfStream; end

  def initialize(*targets)
    @targets = targets
  end

  def each
    return self if @targets.empty?

    threads = ThreadGroup::new

    begin
      main = Thread::start do
	main = Thread::current
	threads.add main
	
	queues = []
	
	@targets.each do |target|
	  queue = SizedQueue::new(1)
	  queues << queue

	  Thread::start(target, queue) do |target, queue|
	    begin
	      target.each do |i|
		queue.push i
	      end
	      queue.push EndOfStream
	    rescue Exception
	      main.raise $!
	    end
	  end
	end
	
	loop do
	  args = queues.collect{|queue| queue.pop}
	  break if args.include? EndOfStream
	  yield args
	end
      end
      
      main.join

    ensure
      threads.list.each do |thread|
	thread.kill 
      end
    end

    self
  end
end


class Array
  def synchronous_each(&block)
    EachSynchronizer::new(*self).each(&block)
  end
end


if __FILE__ == $0
  sample = <<-'END'
  a = [58, 38,  6, 97, 26, 86]
  b = "5\n19\n61\n60\n9\n30"
  c = 50..60
  prog = [a, b, c]
 
  result = []
  
  prog.synchronous_each do |args|
    result << args.collect{|n| n.to_i}.max
  end
  
  p result
  END
  print sample, '=> '
  eval sample
end


=begin

== History

=== 20001/1/23 version 0.1.0

first release.

=end
